#ifndef WRENCHBOARD_API_WRENCHBOARD_H
#define WRENCHBOARD_API_WRENCHBOARD_H

#include "vars.h"
#include "clog.h"

// A very simple wrenchboard class
class WrenchBoard {
public:
    WrenchBoard();
    ~WrenchBoard();
    long wrenchboard_api(CVars in, CVars &out);
    const char* cfgReadChar(const char *parameter);
    long cfgReadLong(const char *parameter);
    void logMessage(const char *message);
private:
    FILE* pFile;
    int db;
    int logFile;
};

#endif /* WRENCHBOARD_API_WRENCHBOARD_H */
